# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import logging
import sys

# Disable C0302: Too many lines in module
# pylint: disable=C0302
import time
from enum import Enum
from pathlib import PurePosixPath
from typing import Dict, Tuple, List
import re
import importlib
import fsspec

from cachetools import TTLCache, LRUCache
from fsspec import AbstractFileSystem
from fsspec.implementations.local import LocalFileSystem
from fsspec.implementations.arrow import ArrowFSWrapper
from fsspec.utils import infer_storage_options

from readerwriterlock import rwlock

from gravitino.api.catalog import Catalog
from gravitino.api.credential.credential import Credential
from gravitino.audit.caller_context import CallerContext, CallerContextHolder
from gravitino.audit.fileset_audit_constants import FilesetAuditConstants
from gravitino.audit.fileset_data_operation import FilesetDataOperation
from gravitino.audit.internal_client_type import InternalClientType
from gravitino.auth.default_oauth2_token_provider import DefaultOAuth2TokenProvider
from gravitino.auth.oauth2_token_provider import OAuth2TokenProvider
from gravitino.auth.simple_auth_provider import SimpleAuthProvider
from gravitino.client.generic_fileset import GenericFileset
from gravitino.client.fileset_catalog import FilesetCatalog
from gravitino.client.gravitino_client import GravitinoClient
from gravitino.exceptions.base import (
    GravitinoRuntimeException,
)
from gravitino.filesystem.gvfs_config import GVFSConfig
from gravitino.name_identifier import NameIdentifier

from gravitino.api.credential.adls_token_credential import ADLSTokenCredential
from gravitino.api.credential.azure_account_key_credential import (
    AzureAccountKeyCredential,
)
from gravitino.api.credential.gcs_token_credential import GCSTokenCredential
from gravitino.api.credential.oss_secret_key_credential import OSSSecretKeyCredential
from gravitino.api.credential.oss_token_credential import OSSTokenCredential
from gravitino.api.credential.s3_secret_key_credential import S3SecretKeyCredential
from gravitino.api.credential.s3_token_credential import S3TokenCredential

logger = logging.getLogger(__name__)

PROTOCOL_NAME = "gvfs"

TIME_WITHOUT_EXPIRATION = sys.maxsize


class StorageType(Enum):
    HDFS = "hdfs"
    LOCAL = "file"
    GCS = "gs"
    S3A = "s3a"
    OSS = "oss"
    ABS = "abfss"


class FilesetContextPair:
    """A context object that holds the information about the actual file location and the file system which used in
    the GravitinoVirtualFileSystem's operations.
    """

    def __init__(self, actual_file_location: str, filesystem: AbstractFileSystem):
        self._actual_file_location = actual_file_location
        self._filesystem = filesystem

    def actual_file_location(self):
        return self._actual_file_location

    def filesystem(self):
        return self._filesystem


class GravitinoVirtualFileSystem(fsspec.AbstractFileSystem):
    """This is a virtual file system that users can access `fileset` and
    other resources.

    It obtains the actual storage location corresponding to the resource from the
    Gravitino server, and creates an independent file system for it to act as an agent for users to
    access the underlying storage.
    """

    # Override the parent variable
    protocol = PROTOCOL_NAME
    _identifier_pattern = re.compile("^fileset/([^/]+)/([^/]+)/([^/]+)(?:/[^/]+)*/?$")
    SLASH = "/"

    def __init__(
        self,
        server_uri: str = None,
        metalake_name: str = None,
        options: Dict = None,
        **kwargs,
    ):
        """Initialize the GravitinoVirtualFileSystem.
        :param server_uri: Gravitino server URI
        :param metalake_name: Gravitino metalake name
        :param options: Options for the GravitinoVirtualFileSystem
        :param kwargs: Extra args for super filesystem
        """
        self._metalake = metalake_name
        auth_type = (
            GVFSConfig.SIMPLE_AUTH_TYPE
            if options is None
            else options.get(GVFSConfig.AUTH_TYPE, GVFSConfig.SIMPLE_AUTH_TYPE)
        )
        if auth_type == GVFSConfig.SIMPLE_AUTH_TYPE:
            self._client = GravitinoClient(
                uri=server_uri,
                metalake_name=metalake_name,
                auth_data_provider=SimpleAuthProvider(),
            )
        elif auth_type == GVFSConfig.OAUTH2_AUTH_TYPE:
            oauth2_server_uri = options.get(GVFSConfig.OAUTH2_SERVER_URI)
            self._check_auth_config(
                auth_type, GVFSConfig.OAUTH2_SERVER_URI, oauth2_server_uri
            )

            oauth2_credential = options.get(GVFSConfig.OAUTH2_CREDENTIAL)
            self._check_auth_config(
                auth_type, GVFSConfig.OAUTH2_CREDENTIAL, oauth2_credential
            )

            oauth2_path = options.get(GVFSConfig.OAUTH2_PATH)
            self._check_auth_config(auth_type, GVFSConfig.OAUTH2_PATH, oauth2_path)

            oauth2_scope = options.get(GVFSConfig.OAUTH2_SCOPE)
            self._check_auth_config(auth_type, GVFSConfig.OAUTH2_SCOPE, oauth2_scope)

            oauth2_token_provider: OAuth2TokenProvider = DefaultOAuth2TokenProvider(
                oauth2_server_uri, oauth2_credential, oauth2_path, oauth2_scope
            )
            self._client = GravitinoClient(
                uri=server_uri,
                metalake_name=metalake_name,
                auth_data_provider=oauth2_token_provider,
            )
        else:
            raise GravitinoRuntimeException(
                f"Authentication type {auth_type} is not supported."
            )
        cache_size = (
            GVFSConfig.DEFAULT_CACHE_SIZE
            if options is None
            else options.get(GVFSConfig.CACHE_SIZE, GVFSConfig.DEFAULT_CACHE_SIZE)
        )
        cache_expired_time = (
            GVFSConfig.DEFAULT_CACHE_EXPIRED_TIME
            if options is None
            else options.get(
                GVFSConfig.CACHE_EXPIRED_TIME, GVFSConfig.DEFAULT_CACHE_EXPIRED_TIME
            )
        )
        self._cache = TTLCache(maxsize=cache_size, ttl=cache_expired_time)
        self._cache_lock = rwlock.RWLockFair()
        self._catalog_cache = LRUCache(maxsize=100)
        self._catalog_cache_lock = rwlock.RWLockFair()
        self._options = options

        super().__init__(**kwargs)

    @property
    def cache(self):
        return self._cache

    @property
    def fsid(self):
        return PROTOCOL_NAME

    def sign(self, path, expiration=None, **kwargs):
        """We do not support to create a signed URL representing the given path in gvfs."""
        raise GravitinoRuntimeException(
            "Sign is not implemented for Gravitino Virtual FileSystem."
        )

    def ls(self, path, detail=True, **kwargs):
        """List the files and directories info of the path.
        :param path: Virtual fileset path
        :param detail: Whether to show the details for the files and directories info
        :param kwargs: Extra args
        :return If details is true, returns a list of file info dicts, else returns a list of file paths
        """
        context_pair: FilesetContextPair = self._get_fileset_context(
            path, FilesetDataOperation.LIST_STATUS
        )
        actual_path = context_pair.actual_file_location()
        storage_type = self._recognize_storage_type(actual_path)
        pre_process_path: str = self._pre_process_path(path)
        identifier: NameIdentifier = self._extract_identifier(pre_process_path)
        sub_path: str = self._get_sub_path_from_virtual_path(
            identifier, pre_process_path
        )
        storage_location: str = actual_path[: len(actual_path) - len(sub_path)]
        # return entries with details
        if detail:
            entries = context_pair.filesystem().ls(
                self._strip_storage_protocol(storage_type, actual_path),
                detail=True,
            )
            virtual_entries = [
                self._convert_actual_info(
                    entry, storage_location, self._get_virtual_location(identifier)
                )
                for entry in entries
            ]
            return virtual_entries
        # only returns paths
        entry_paths = context_pair.filesystem().ls(
            self._strip_storage_protocol(storage_type, actual_path),
            detail=False,
        )
        virtual_entry_paths = [
            self._convert_actual_path(
                entry_path, storage_location, self._get_virtual_location(identifier)
            )
            for entry_path in entry_paths
        ]
        return virtual_entry_paths

    def info(self, path, **kwargs):
        """Get file info.
        :param path: Virtual fileset path
        :param kwargs: Extra args
        :return A file info dict
        """
        context_pair: FilesetContextPair = self._get_fileset_context(
            path, FilesetDataOperation.GET_FILE_STATUS
        )
        actual_path = context_pair.actual_file_location()
        storage_type = self._recognize_storage_type(actual_path)
        pre_process_path: str = self._pre_process_path(path)
        identifier: NameIdentifier = self._extract_identifier(pre_process_path)
        sub_path: str = self._get_sub_path_from_virtual_path(
            identifier, pre_process_path
        )
        storage_location: str = actual_path[: len(actual_path) - len(sub_path)]
        actual_info: Dict = context_pair.filesystem().info(
            self._strip_storage_protocol(storage_type, actual_path)
        )
        return self._convert_actual_info(
            actual_info, storage_location, self._get_virtual_location(identifier)
        )

    def exists(self, path, **kwargs):
        """Check if a file or a directory exists.
        :param path: Virtual fileset path
        :param kwargs: Extra args
        :return If a file or directory exists, it returns True, otherwise False
        """
        context_pair: FilesetContextPair = self._get_fileset_context(
            path, FilesetDataOperation.EXISTS
        )
        actual_path = context_pair.actual_file_location()
        storage_type = self._recognize_storage_type(actual_path)
        return context_pair.filesystem().exists(
            self._strip_storage_protocol(storage_type, actual_path)
        )

    def cp_file(self, path1, path2, **kwargs):
        """Copy a file.
        :param path1: Virtual src fileset path
        :param path2: Virtual dst fileset path, should be consistent with the src path fileset identifier
        :param kwargs: Extra args
        """
        src_path = self._pre_process_path(path1)
        dst_path = self._pre_process_path(path2)
        src_identifier: NameIdentifier = self._extract_identifier(src_path)
        dst_identifier: NameIdentifier = self._extract_identifier(dst_path)
        if src_identifier != dst_identifier:
            raise GravitinoRuntimeException(
                f"Destination file path identifier: `{dst_identifier}` should be same with src file path "
                f"identifier: `{src_identifier}`."
            )
        src_context_pair: FilesetContextPair = self._get_fileset_context(
            src_path, FilesetDataOperation.COPY_FILE
        )
        src_actual_path = src_context_pair.actual_file_location()

        dst_context_pair: FilesetContextPair = self._get_fileset_context(
            dst_path, FilesetDataOperation.COPY_FILE
        )
        dst_actual_path = dst_context_pair.actual_file_location()

        storage_type = self._recognize_storage_type(src_actual_path)
        src_context_pair.filesystem().cp_file(
            self._strip_storage_protocol(storage_type, src_actual_path),
            self._strip_storage_protocol(storage_type, dst_actual_path),
        )

    def mv(self, path1, path2, recursive=False, maxdepth=None, **kwargs):
        """Move a file to another directory.
         This can move a file to another existing directory.
         If the target path directory does not exist, an exception will be thrown.
        :param path1: Virtual src fileset path
        :param path2: Virtual dst fileset path, should be consistent with the src path fileset identifier
        :param recursive: Whether to move recursively
        :param maxdepth: Maximum depth of recursive move
        :param kwargs: Extra args
        """
        src_path = self._pre_process_path(path1)
        dst_path = self._pre_process_path(path2)
        src_identifier: NameIdentifier = self._extract_identifier(src_path)
        dst_identifier: NameIdentifier = self._extract_identifier(dst_path)
        if src_identifier != dst_identifier:
            raise GravitinoRuntimeException(
                f"Destination file path identifier: `{dst_identifier}`"
                f" should be same with src file path identifier: `{src_identifier}`."
            )
        src_context_pair: FilesetContextPair = self._get_fileset_context(
            src_path, FilesetDataOperation.RENAME
        )
        src_actual_path = src_context_pair.actual_file_location()
        storage_type = self._recognize_storage_type(src_actual_path)
        dst_context_pair: FilesetContextPair = self._get_fileset_context(
            dst_path, FilesetDataOperation.RENAME
        )
        dst_actual_path = dst_context_pair.actual_file_location()

        # convert the following to in

        if storage_type in [
            StorageType.HDFS,
            StorageType.GCS,
            StorageType.S3A,
            StorageType.OSS,
            StorageType.ABS,
        ]:
            src_context_pair.filesystem().mv(
                self._strip_storage_protocol(storage_type, src_actual_path),
                self._strip_storage_protocol(storage_type, dst_actual_path),
            )
        elif storage_type == StorageType.LOCAL:
            src_context_pair.filesystem().mv(
                self._strip_storage_protocol(storage_type, src_actual_path),
                self._strip_storage_protocol(storage_type, dst_actual_path),
                recursive,
                maxdepth,
            )
        else:
            raise GravitinoRuntimeException(
                f"Storage type:{storage_type} doesn't support now."
            )

    def _rm(self, path):
        raise GravitinoRuntimeException(
            "Deprecated method, use `rm_file` method instead."
        )

    def lazy_load_class(self, module_name, class_name):
        module = importlib.import_module(module_name)
        return getattr(module, class_name)

    def rm(self, path, recursive=False, maxdepth=None):
        """Remove a file or directory.
        :param path: Virtual fileset path
        :param recursive: Whether to remove the directory recursively.
                When removing a directory, this parameter should be True.
        :param maxdepth: The maximum depth to remove the directory recursively.
        """
        context_pair: FilesetContextPair = self._get_fileset_context(
            path, FilesetDataOperation.DELETE
        )
        actual_path = context_pair.actual_file_location()
        storage_type = self._recognize_storage_type(actual_path)
        fs = context_pair.filesystem()

        # S3FileSystem doesn't support maxdepth
        if isinstance(fs, self.lazy_load_class("s3fs", "S3FileSystem")):
            fs.rm(self._strip_storage_protocol(storage_type, actual_path), recursive)
        else:
            fs.rm(
                self._strip_storage_protocol(storage_type, actual_path),
                recursive,
                maxdepth,
            )

    def rm_file(self, path):
        """Remove a file.
        :param path: Virtual fileset path
        """
        context_pair: FilesetContextPair = self._get_fileset_context(
            path, FilesetDataOperation.DELETE
        )
        actual_path = context_pair.actual_file_location()
        storage_type = self._recognize_storage_type(actual_path)
        context_pair.filesystem().rm_file(
            self._strip_storage_protocol(storage_type, actual_path)
        )

    def rmdir(self, path):
        """Remove a directory.
        It will delete a directory and all its contents recursively for PyArrow.HadoopFileSystem.
        And it will throw an exception if delete a directory which is non-empty for LocalFileSystem.
        :param path: Virtual fileset path
        """
        context_pair: FilesetContextPair = self._get_fileset_context(
            path, FilesetDataOperation.DELETE
        )
        actual_path = context_pair.actual_file_location()
        storage_type = self._recognize_storage_type(actual_path)
        context_pair.filesystem().rmdir(
            self._strip_storage_protocol(storage_type, actual_path)
        )

    def open(
        self,
        path,
        mode="rb",
        block_size=None,
        cache_options=None,
        compression=None,
        **kwargs,
    ):
        """Open a file to read/write/append.
        :param path: Virtual fileset path
        :param mode: The mode now supports: rb(read), wb(write), ab(append). See builtin ``open()``
        :param block_size: Some indication of buffering - this is a value in bytes
        :param cache_options: Extra arguments to pass through to the cache
        :param compression: If given, open file using compression codec
        :param kwargs: Extra args
        :return A file-like object from the filesystem
        """
        if mode in ("w", "wb"):
            data_operation = FilesetDataOperation.OPEN_AND_WRITE
        elif mode in ("a", "ab"):
            data_operation = FilesetDataOperation.OPEN_AND_APPEND
        else:
            data_operation = FilesetDataOperation.OPEN
        context_pair: FilesetContextPair = self._get_fileset_context(
            path, data_operation
        )
        actual_path = context_pair.actual_file_location()
        storage_type = self._recognize_storage_type(actual_path)
        return context_pair.filesystem().open(
            self._strip_storage_protocol(storage_type, actual_path),
            mode,
            block_size,
            cache_options,
            compression,
            **kwargs,
        )

    def mkdir(self, path, create_parents=True, **kwargs):
        """Make a directory.
        if create_parents=True, this is equivalent to ``makedirs``.

        :param path: Virtual fileset path
        :param create_parents: Create parent directories if missing when set to True
        :param kwargs: Extra args
        """
        context_pair: FilesetContextPair = self._get_fileset_context(
            path, FilesetDataOperation.MKDIRS
        )
        actual_path = context_pair.actual_file_location()
        storage_type = self._recognize_storage_type(actual_path)
        context_pair.filesystem().mkdir(
            self._strip_storage_protocol(storage_type, actual_path),
            create_parents,
            **kwargs,
        )

    def makedirs(self, path, exist_ok=True):
        """Make a directory recursively.
        :param path: Virtual fileset path
        :param exist_ok: Continue if a directory already exists
        """
        context_pair: FilesetContextPair = self._get_fileset_context(
            path, FilesetDataOperation.MKDIRS
        )
        actual_path = context_pair.actual_file_location()
        storage_type = self._recognize_storage_type(actual_path)
        context_pair.filesystem().makedirs(
            self._strip_storage_protocol(storage_type, actual_path),
            exist_ok,
        )

    def created(self, path):
        """Return the created timestamp of a file as a datetime.datetime
        Only supports for `fsspec.LocalFileSystem` now.
        :param path: Virtual fileset path
        :return Created time(datetime.datetime)
        """
        context_pair: FilesetContextPair = self._get_fileset_context(
            path, FilesetDataOperation.CREATED_TIME
        )
        actual_path = context_pair.actual_file_location()
        storage_type = self._recognize_storage_type(actual_path)
        if storage_type == StorageType.LOCAL:
            return context_pair.filesystem().created(
                self._strip_storage_protocol(storage_type, actual_path)
            )
        raise GravitinoRuntimeException(
            f"Storage type:{storage_type} doesn't support now."
        )

    def modified(self, path):
        """Returns the modified time of the path file if it exists.
        :param path: Virtual fileset path
        :return Modified time(datetime.datetime)
        """
        context_pair: FilesetContextPair = self._get_fileset_context(
            path, FilesetDataOperation.MODIFIED_TIME
        )
        actual_path = context_pair.actual_file_location()
        storage_type = self._recognize_storage_type(actual_path)
        return context_pair.filesystem().modified(
            self._strip_storage_protocol(storage_type, actual_path)
        )

    def cat_file(self, path, start=None, end=None, **kwargs):
        """Get the content of a file.
        :param path: Virtual fileset path
        :param start: The offset in bytes to start reading from. It can be None.
        :param end: The offset in bytes to end reading at. It can be None.
        :param kwargs: Extra args
        :return File content
        """
        context_pair: FilesetContextPair = self._get_fileset_context(
            path, FilesetDataOperation.CAT_FILE
        )
        actual_path = context_pair.actual_file_location()
        storage_type = self._recognize_storage_type(actual_path)
        return context_pair.filesystem().cat_file(
            self._strip_storage_protocol(storage_type, actual_path),
            start,
            end,
            **kwargs,
        )

    def get_file(self, rpath, lpath, callback=None, outfile=None, **kwargs):
        """Copy single remote file to local.
        :param rpath: Remote file path
        :param lpath: Local file path
        :param callback: The callback class
        :param outfile: The output file path
        :param kwargs: Extra args
        """
        if not lpath.startswith(f"{StorageType.LOCAL.value}:") and not lpath.startswith(
            "/"
        ):
            raise GravitinoRuntimeException(
                "Doesn't support copy a remote gvfs file to an another remote file."
            )
        context_pair: FilesetContextPair = self._get_fileset_context(
            rpath, FilesetDataOperation.GET_FILE
        )
        actual_path = context_pair.actual_file_location()
        storage_type = self._recognize_storage_type(actual_path)
        context_pair.filesystem().get_file(
            self._strip_storage_protocol(storage_type, actual_path),
            lpath,
            **kwargs,
        )

    def _convert_actual_path(
        self,
        actual_path: str,
        storage_location: str,
        virtual_location: str,
    ):
        """Convert an actual path to a virtual path.
          The virtual path is like `fileset/{catalog}/{schema}/{fileset}/xxx`.
        :param actual_path: Actual path
        :param storage_location: Storage location
        :param virtual_location: Virtual location
        :return A virtual path
        """

        # If the storage path starts with hdfs, gcs, we should use the path as the prefix.
        if (
            storage_location.startswith(f"{StorageType.HDFS.value}://")
            or storage_location.startswith(f"{StorageType.GCS.value}://")
            or storage_location.startswith(f"{StorageType.S3A.value}://")
        ):
            actual_prefix = infer_storage_options(storage_location)["path"]
        elif storage_location.startswith(f"{StorageType.OSS.value}:/"):
            ops = infer_storage_options(storage_location)
            if "host" not in ops or "path" not in ops:
                raise GravitinoRuntimeException(
                    f"Storage location:{storage_location} doesn't support now."
                )

            actual_prefix = ops["host"] + ops["path"]
        elif storage_location.startswith(f"{StorageType.ABS.value}://"):
            ops = infer_storage_options(storage_location)
            if "username" not in ops or "host" not in ops or "path" not in ops:
                raise GravitinoRuntimeException(
                    f"Storage location:{storage_location} doesn't support now, the username,"
                    f"host and path are required in the storage location."
                )
            actual_prefix = f"{StorageType.ABS.value}://{ops['username']}@{ops['host']}{ops['path']}"

            # the actual path may be '{container}/{path}', we need to add the host and username
            # get the path from {container}/{path}
            if not actual_path.startswith(f"{StorageType.ABS}"):
                path_without_username = actual_path[actual_path.index("/") + 1 :]
                actual_path = f"{StorageType.ABS.value}://{ops['username']}@{ops['host']}/{path_without_username}"

        elif storage_location.startswith(f"{StorageType.LOCAL.value}:/"):
            actual_prefix = storage_location[len(f"{StorageType.LOCAL.value}:") :]
        else:
            raise GravitinoRuntimeException(
                f"Storage location:{storage_location} doesn't support now."
            )

        if not actual_path.startswith(actual_prefix):
            raise GravitinoRuntimeException(
                f"Path {actual_path} does not start with valid prefix {actual_prefix}."
            )

        # if the storage location is end with "/",
        # we should truncate this to avoid replace issues.
        if actual_prefix.endswith(self.SLASH) and not virtual_location.endswith(
            self.SLASH
        ):
            return f"{actual_path.replace(actual_prefix[:-1], virtual_location)}"
        return f"{actual_path.replace(actual_prefix, virtual_location)}"

    def _convert_actual_info(
        self,
        entry: Dict,
        storage_location: str,
        virtual_location: str,
    ):
        """Convert a file info from an actual entry to a virtual entry.
        :param entry: A dict of the actual file info
        :param storage_location: Storage location
        :param virtual_location: Virtual location
        :return A dict of the virtual file info
        """
        path = self._convert_actual_path(
            entry["name"], storage_location, virtual_location
        )

        last_modified = None
        if "mtime" in entry:
            # HDFS and GCS
            last_modified = entry["mtime"]
        elif "LastModified" in entry:
            # S3 and OSS
            last_modified = entry["LastModified"]
        elif "last_modified" in entry:
            # Azure Blob Storage
            last_modified = entry["last_modified"]

        return {
            "name": path,
            "size": entry["size"],
            "type": entry["type"],
            "mtime": last_modified,
        }

    def _get_fileset_context(self, virtual_path: str, operation: FilesetDataOperation):
        """Get a fileset context from the cache or the Gravitino server
        :param virtual_path: The virtual path
        :param operation: The data operation
        :return A fileset context pair
        """
        virtual_path: str = self._pre_process_path(virtual_path)
        identifier: NameIdentifier = self._extract_identifier(virtual_path)
        catalog_ident: NameIdentifier = NameIdentifier.of(
            self._metalake, identifier.namespace().level(1)
        )
        fileset_catalog = self._get_fileset_catalog(catalog_ident)
        if fileset_catalog is None:
            raise GravitinoRuntimeException(
                f"Loaded fileset catalog: {catalog_ident} is null."
            )
        sub_path: str = self._get_sub_path_from_virtual_path(identifier, virtual_path)
        context = {
            FilesetAuditConstants.HTTP_HEADER_FILESET_DATA_OPERATION: operation.name,
            FilesetAuditConstants.HTTP_HEADER_INTERNAL_CLIENT_TYPE: InternalClientType.PYTHON_GVFS.name,
        }
        caller_context: CallerContext = CallerContext(context)
        CallerContextHolder.set(caller_context)
        actual_file_location: (
            str
        ) = fileset_catalog.as_fileset_catalog().get_file_location(
            NameIdentifier.of(identifier.namespace().level(2), identifier.name()),
            sub_path,
        )

        return FilesetContextPair(
            actual_file_location,
            self._get_filesystem(actual_file_location, fileset_catalog, identifier),
        )

    def _extract_identifier(self, path):
        """Extract the fileset identifier from the path.
        :param path: The virtual fileset path
        :return The fileset identifier
        """
        if path is None:
            raise GravitinoRuntimeException(
                "path which need be extracted cannot be null or empty."
            )

        match = self._identifier_pattern.match(path)
        if match and len(match.groups()) == 3:
            return NameIdentifier.of(
                self._metalake, match.group(1), match.group(2), match.group(3)
            )
        raise GravitinoRuntimeException(
            f"path: `{path}` doesn't contains valid identifier."
        )

    @staticmethod
    def _get_virtual_location(identifier: NameIdentifier):
        """Get the virtual location of the fileset.
        :param identifier: The name identifier of the fileset
        :return The virtual location.
        """
        return (
            f"fileset/{identifier.namespace().level(1)}"
            f"/{identifier.namespace().level(2)}"
            f"/{identifier.name()}"
        )

    @staticmethod
    def _pre_process_path(virtual_path):
        """Pre-process the path.
         We will uniformly process `gvfs://fileset/{catalog}/{schema}/{fileset_name}/xxx`
         into the format of `fileset/{catalog}/{schema}/{fileset_name}/xxx`.
         This is because some implementations of `PyArrow` and `fsspec` can only recognize this format.
        :param virtual_path: The virtual path
        :return The pre-processed path
        """
        if isinstance(virtual_path, PurePosixPath):
            pre_processed_path = virtual_path.as_posix()
        else:
            pre_processed_path = virtual_path
        gvfs_prefix = f"{PROTOCOL_NAME}://"
        if pre_processed_path.startswith(gvfs_prefix):
            pre_processed_path = pre_processed_path[len(gvfs_prefix) :]
        if not pre_processed_path.startswith("fileset/"):
            raise GravitinoRuntimeException(
                f"Invalid path:`{pre_processed_path}`. Expected path to start with `fileset/`."
                " Example: fileset/{fileset_catalog}/{schema}/{fileset_name}/{sub_path}."
            )
        return pre_processed_path

    @staticmethod
    def _recognize_storage_type(path: str):
        """Recognize the storage type by the path.
        :param path: The path
        :return: The storage type
        """
        if path.startswith(f"{StorageType.HDFS.value}://"):
            return StorageType.HDFS
        if path.startswith(f"{StorageType.LOCAL.value}:/"):
            return StorageType.LOCAL
        if path.startswith(f"{StorageType.GCS.value}://"):
            return StorageType.GCS
        if path.startswith(f"{StorageType.S3A.value}://"):
            return StorageType.S3A
        if path.startswith(f"{StorageType.OSS.value}://"):
            return StorageType.OSS
        if path.startswith(f"{StorageType.ABS.value}://"):
            return StorageType.ABS
        raise GravitinoRuntimeException(
            f"Storage type doesn't support now. Path:{path}"
        )

    @staticmethod
    def _get_sub_path_from_virtual_path(identifier: NameIdentifier, virtual_path: str):
        return virtual_path[
            len(
                f"fileset/{identifier.namespace().level(1)}/{identifier.namespace().level(2)}/{identifier.name()}"
            ) :
        ]

    @staticmethod
    def _strip_storage_protocol(storage_type: StorageType, path: str):
        """Strip the storage protocol from the path.
          Before passing the path to the underlying file system for processing,
           pre-process the protocol information in the path.
          Some file systems require special processing.
          For HDFS, we can pass the path like 'hdfs://{host}:{port}/xxx'.
          For Local, we can pass the path like '/tmp/xxx'.
        :param storage_type: The storage type
        :param path: The path
        :return: The stripped path

        We will handle OSS differently from S3 and GCS, because OSS has different behavior than S3 and GCS.
        Please see the following example:

        ```
        >> oss = context_pair.filesystem()
        >> oss.ls('oss://bucket-xiaoyu/test_gvfs_catalog678/test_gvfs_schema/test_gvfs_fileset/test_ls')
            DEBUG:ossfs:Get directory listing page for bucket-xiaoyu/test_gvfs_catalog678/
            test_gvfs_schema/test_gvfs_fileset
            DEBUG:ossfs:CALL: ObjectIterator - () - {'prefix': 'test_gvfs_catalog678/test_gvfs_schema
            /test_gvfs_fileset/', 'delimiter': '/'}
            []
        >> oss.ls('bucket-xiaoyu/test_gvfs_catalog678/test_gvfs_schema/test_gvfs_fileset/test_ls')
            DEBUG:ossfs:Get directory listing page for bucket-xiaoyu/test_gvfs_catalog678/test_gvfs_schema
            /test_gvfs_fileset/test_ls
            DEBUG:ossfs:CALL: ObjectIterator - () - {'prefix': 'test_gvfs_catalog678/test_gvfs_schema
            /test_gvfs_fileset/test_ls/', 'delimiter': '/'}
            [{'name': 'bucket-xiaoyu/test_gvfs_catalog678/test_gvfs_schema/test_gvfs_fileset/test_ls
            /test.file', 'type': 'file', 'size': 0, 'LastModified': 1729754793,
            'Size': 0, 'Key': 'bucket-xiaoyu/test_gvfs_catalog678/test_gvfs_schema/
            test_gvfs_fileset/test_ls/test.file'}]

        ```

        Please take a look at the above example: if we do not remove the protocol (starts with oss://),
        it will always return an empty array when we call `oss.ls`, however, if we remove the protocol,
        it will produce the correct result as expected.
        """
        if storage_type in (StorageType.HDFS, StorageType.GCS, StorageType.S3A):
            return path
        if storage_type == StorageType.LOCAL:
            return path[len(f"{StorageType.LOCAL.value}:") :]

        ## We need to remove the protocol and accout from the path, for instance,
        # the path can be converted from 'abfss://container@account/path' to
        # 'container/path'.
        if storage_type == StorageType.ABS:
            ops = infer_storage_options(path)
            return ops["username"] + ops["path"]

        # OSS has different behavior than S3 and GCS, if we do not remove the
        # protocol, it will always return an empty array.
        if storage_type == StorageType.OSS:
            if path.startswith(f"{StorageType.OSS.value}://"):
                return path[len(f"{StorageType.OSS.value}://") :]
            return path

        raise GravitinoRuntimeException(
            f"Storage type:{storage_type} doesn't support now."
        )

    @staticmethod
    def _check_auth_config(auth_type: str, config_key: str, config_value: str):
        """Check if the config value is null.
        :param auth_type: The auth type
        :param config_key: The config key
        :param config_value: The config value
        """
        if config_value is None:
            raise GravitinoRuntimeException(
                f"{config_key} should not be null"
                f" if {GVFSConfig.AUTH_TYPE} is set to {auth_type}."
            )

    def _get_fileset_catalog(self, catalog_ident: NameIdentifier):
        read_lock = self._catalog_cache_lock.gen_rlock()
        try:
            read_lock.acquire()
            cache_value: Tuple[NameIdentifier, FilesetCatalog] = (
                self._catalog_cache.get(catalog_ident)
            )
            if cache_value is not None:
                return cache_value
        finally:
            read_lock.release()

        write_lock = self._catalog_cache_lock.gen_wlock()
        try:
            write_lock.acquire()
            cache_value: Tuple[NameIdentifier, FilesetCatalog] = (
                self._catalog_cache.get(catalog_ident)
            )
            if cache_value is not None:
                return cache_value
            catalog = self._client.load_catalog(catalog_ident.name())
            self._catalog_cache[catalog_ident] = catalog
            return catalog
        finally:
            write_lock.release()

    def _file_system_expired(self, expire_time: int):
        return expire_time <= time.time() * 1000

    # Disable Too many branches (13/12) (too-many-branches)
    # pylint: disable=R0912
    def _get_filesystem(
        self,
        actual_file_location: str,
        fileset_catalog: Catalog,
        name_identifier: NameIdentifier,
    ):
        storage_type = self._recognize_storage_type(actual_file_location)
        read_lock = self._cache_lock.gen_rlock()
        try:
            read_lock.acquire()
            cache_value: Tuple[int, AbstractFileSystem] = self._cache.get(
                name_identifier
            )
            if cache_value is not None:
                if not self._file_system_expired(cache_value[0]):
                    return cache_value[1]
        finally:
            read_lock.release()

        write_lock = self._cache_lock.gen_wlock()
        try:
            write_lock.acquire()
            cache_value: Tuple[int, AbstractFileSystem] = self._cache.get(
                name_identifier
            )

            if cache_value is not None:
                if not self._file_system_expired(cache_value[0]):
                    return cache_value[1]

            new_cache_value: Tuple[int, AbstractFileSystem]
            if storage_type == StorageType.HDFS:
                fs_class = importlib.import_module("pyarrow.fs").HadoopFileSystem
                new_cache_value = (
                    TIME_WITHOUT_EXPIRATION,
                    ArrowFSWrapper(fs_class.from_uri(actual_file_location)),
                )
            elif storage_type == StorageType.LOCAL:
                new_cache_value = (TIME_WITHOUT_EXPIRATION, LocalFileSystem())
            elif storage_type == StorageType.GCS:
                new_cache_value = self._get_gcs_filesystem(
                    fileset_catalog, name_identifier
                )
            elif storage_type == StorageType.S3A:
                new_cache_value = self._get_s3_filesystem(
                    fileset_catalog, name_identifier
                )
            elif storage_type == StorageType.OSS:
                new_cache_value = self._get_oss_filesystem(
                    fileset_catalog, name_identifier
                )
            elif storage_type == StorageType.ABS:
                new_cache_value = self._get_abs_filesystem(
                    fileset_catalog, name_identifier
                )
            else:
                raise GravitinoRuntimeException(
                    f"Storage type: `{storage_type}` doesn't support now."
                )
            self._cache[name_identifier] = new_cache_value
            return new_cache_value[1]
        finally:
            write_lock.release()

    def _get_gcs_filesystem(self, fileset_catalog: Catalog, identifier: NameIdentifier):
        fileset: GenericFileset = fileset_catalog.as_fileset_catalog().load_fileset(
            NameIdentifier.of(identifier.namespace().level(2), identifier.name())
        )
        credentials = fileset.support_credentials().get_credentials()

        credential = self._get_most_suitable_gcs_credential(credentials)
        if credential is not None:
            expire_time = self._get_expire_time_by_ratio(credential.expire_time_in_ms())
            if isinstance(credential, GCSTokenCredential):
                fs = importlib.import_module("gcsfs").GCSFileSystem(
                    token=credential.token()
                )
                return (expire_time, fs)

        # get 'service-account-key' from gcs_options, if the key is not found, throw an exception
        service_account_key_path = self._options.get(
            GVFSConfig.GVFS_FILESYSTEM_GCS_SERVICE_KEY_FILE
        )
        if service_account_key_path is None:
            raise GravitinoRuntimeException(
                "Service account key is not found in the options."
            )
        return (
            TIME_WITHOUT_EXPIRATION,
            importlib.import_module("gcsfs").GCSFileSystem(
                token=service_account_key_path
            ),
        )

    def _get_s3_filesystem(self, fileset_catalog: Catalog, identifier: NameIdentifier):
        fileset: GenericFileset = fileset_catalog.as_fileset_catalog().load_fileset(
            NameIdentifier.of(identifier.namespace().level(2), identifier.name())
        )
        credentials = fileset.support_credentials().get_credentials()
        credential = self._get_most_suitable_s3_credential(credentials)

        # S3 endpoint from gravitino server, Note: the endpoint may not a real S3 endpoint
        # it can be a simulated S3 endpoint, such as minio, so though the endpoint is not a required field
        # for S3FileSystem, we still need to assign the endpoint to the S3FileSystem
        s3_endpoint = fileset_catalog.properties().get("s3-endpoint", None)
        # If the oss endpoint is not found in the fileset catalog, get it from the client options
        if s3_endpoint is None:
            s3_endpoint = self._options.get(GVFSConfig.GVFS_FILESYSTEM_S3_ENDPOINT)

        if credential is not None:
            expire_time = self._get_expire_time_by_ratio(credential.expire_time_in_ms())
            if isinstance(credential, S3TokenCredential):
                fs = importlib.import_module("s3fs").S3FileSystem(
                    key=credential.access_key_id(),
                    secret=credential.secret_access_key(),
                    token=credential.session_token(),
                    endpoint_url=s3_endpoint,
                )
                return (expire_time, fs)
            if isinstance(credential, S3SecretKeyCredential):
                fs = importlib.import_module("s3fs").S3FileSystem(
                    key=credential.access_key_id(),
                    secret=credential.secret_access_key(),
                    endpoint_url=s3_endpoint,
                )
                return (expire_time, fs)

        # this is the old way to get the s3 file system
        # get 'aws_access_key_id' from s3_options, if the key is not found, throw an exception
        aws_access_key_id = self._options.get(GVFSConfig.GVFS_FILESYSTEM_S3_ACCESS_KEY)
        if aws_access_key_id is None:
            raise GravitinoRuntimeException(
                "AWS access key id is not found in the options."
            )

        # get 'aws_secret_access_key' from s3_options, if the key is not found, throw an exception
        aws_secret_access_key = self._options.get(
            GVFSConfig.GVFS_FILESYSTEM_S3_SECRET_KEY
        )
        if aws_secret_access_key is None:
            raise GravitinoRuntimeException(
                "AWS secret access key is not found in the options."
            )

        return (
            TIME_WITHOUT_EXPIRATION,
            importlib.import_module("s3fs").S3FileSystem(
                key=aws_access_key_id,
                secret=aws_secret_access_key,
                endpoint_url=s3_endpoint,
            ),
        )

    def _get_oss_filesystem(self, fileset_catalog: Catalog, identifier: NameIdentifier):
        fileset: GenericFileset = fileset_catalog.as_fileset_catalog().load_fileset(
            NameIdentifier.of(identifier.namespace().level(2), identifier.name())
        )
        credentials = fileset.support_credentials().get_credentials()

        # OSS endpoint from gravitino server
        oss_endpoint = fileset_catalog.properties().get("oss-endpoint", None)
        # If the oss endpoint is not found in the fileset catalog, get it from the client options
        if oss_endpoint is None:
            oss_endpoint = self._options.get(GVFSConfig.GVFS_FILESYSTEM_OSS_ENDPOINT)

        credential = self._get_most_suitable_oss_credential(credentials)
        if credential is not None:
            expire_time = self._get_expire_time_by_ratio(credential.expire_time_in_ms())
            if isinstance(credential, OSSTokenCredential):
                fs = importlib.import_module("ossfs").OSSFileSystem(
                    key=credential.access_key_id(),
                    secret=credential.secret_access_key(),
                    token=credential.security_token(),
                    endpoint=oss_endpoint,
                )
                return (expire_time, fs)
            if isinstance(credential, OSSSecretKeyCredential):
                return (
                    expire_time,
                    importlib.import_module("ossfs").OSSFileSystem(
                        key=credential.access_key_id(),
                        secret=credential.secret_access_key(),
                        endpoint=oss_endpoint,
                    ),
                )

        # get 'oss_access_key_id' from oss options, if the key is not found, throw an exception
        oss_access_key_id = self._options.get(GVFSConfig.GVFS_FILESYSTEM_OSS_ACCESS_KEY)
        if oss_access_key_id is None:
            raise GravitinoRuntimeException(
                "OSS access key id is not found in the options."
            )

        # get 'oss_secret_access_key' from oss options, if the key is not found, throw an exception
        oss_secret_access_key = self._options.get(
            GVFSConfig.GVFS_FILESYSTEM_OSS_SECRET_KEY
        )
        if oss_secret_access_key is None:
            raise GravitinoRuntimeException(
                "OSS secret access key is not found in the options."
            )

        return (
            TIME_WITHOUT_EXPIRATION,
            importlib.import_module("ossfs").OSSFileSystem(
                key=oss_access_key_id,
                secret=oss_secret_access_key,
                endpoint=oss_endpoint,
            ),
        )

    def _get_abs_filesystem(self, fileset_catalog: Catalog, identifier: NameIdentifier):
        fileset: GenericFileset = fileset_catalog.as_fileset_catalog().load_fileset(
            NameIdentifier.of(identifier.namespace().level(2), identifier.name())
        )
        credentials = fileset.support_credentials().get_credentials()

        credential = self._get_most_suitable_abs_credential(credentials)
        if credential is not None:
            expire_time = self._get_expire_time_by_ratio(credential.expire_time_in_ms())
            if isinstance(credential, ADLSTokenCredential):
                fs = importlib.import_module("adlfs").AzureBlobFileSystem(
                    account_name=credential.account_name(),
                    sas_token=credential.sas_token(),
                )
                return (expire_time, fs)

            if isinstance(credential, AzureAccountKeyCredential):
                fs = importlib.import_module("adlfs").AzureBlobFileSystem(
                    account_name=credential.account_name(),
                    account_key=credential.account_key(),
                )
                return (expire_time, fs)

        # get 'abs_account_name' from abs options, if the key is not found, throw an exception
        abs_account_name = self._options.get(
            GVFSConfig.GVFS_FILESYSTEM_AZURE_ACCOUNT_NAME
        )
        if abs_account_name is None:
            raise GravitinoRuntimeException(
                "ABS account name is not found in the options."
            )

        # get 'abs_account_key' from abs options, if the key is not found, throw an exception
        abs_account_key = self._options.get(
            GVFSConfig.GVFS_FILESYSTEM_AZURE_ACCOUNT_KEY
        )
        if abs_account_key is None:
            raise GravitinoRuntimeException(
                "ABS account key is not found in the options."
            )

        return (
            TIME_WITHOUT_EXPIRATION,
            importlib.import_module("adlfs").AzureBlobFileSystem(
                account_name=abs_account_name,
                account_key=abs_account_key,
            ),
        )

    def _get_most_suitable_s3_credential(self, credentials: List[Credential]):
        for credential in credentials:
            # Prefer to use the token credential, if it does not exist, use the
            # secret key credential.
            if isinstance(credential, S3TokenCredential):
                return credential

        for credential in credentials:
            if isinstance(credential, S3SecretKeyCredential):
                return credential
        return None

    def _get_most_suitable_oss_credential(self, credentials: List[Credential]):
        for credential in credentials:
            # Prefer to use the token credential, if it does not exist, use the
            # secret key credential.
            if isinstance(credential, OSSTokenCredential):
                return credential

        for credential in credentials:
            if isinstance(credential, OSSSecretKeyCredential):
                return credential
        return None

    def _get_most_suitable_gcs_credential(self, credentials: List[Credential]):
        for credential in credentials:
            # Prefer to use the token credential, if it does not exist, return None.
            if isinstance(credential, GCSTokenCredential):
                return credential
        return None

    def _get_most_suitable_abs_credential(self, credentials: List[Credential]):
        for credential in credentials:
            # Prefer to use the token credential, if it does not exist, use the
            # account key credential
            if isinstance(credential, ADLSTokenCredential):
                return credential

        for credential in credentials:
            if isinstance(credential, AzureAccountKeyCredential):
                return credential
        return None

    def _get_expire_time_by_ratio(self, expire_time: int):
        if expire_time <= 0:
            return TIME_WITHOUT_EXPIRATION

        ratio = float(
            self._options.get(
                GVFSConfig.GVFS_FILESYSTEM_CREDENTIAL_EXPIRED_TIME_RATIO,
                GVFSConfig.DEFAULT_CREDENTIAL_EXPIRED_TIME_RATIO,
            )
        )
        return time.time() * 1000 + (expire_time - time.time() * 1000) * ratio


fsspec.register_implementation(PROTOCOL_NAME, GravitinoVirtualFileSystem)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import abstractmethod
from dataclasses import field, dataclass
from typing import Optional

from dataclasses_json import config

from gravitino.api.catalog_change import CatalogChange
from gravitino.rest.rest_message import RESTRequest


@dataclass
class CatalogUpdateRequestBase(RESTRequest):
    _type: str = field(metadata=config(field_name="@type"))

    def __init__(self, action_type: str):
        self._type = action_type

    @abstractmethod
    def catalog_change(self):
        pass


class CatalogUpdateRequest:
    """Represents an interface for catalog update requests."""

    @dataclass
    class RenameCatalogRequest(CatalogUpdateRequestBase):
        """Represents a request to rename a catalog."""

        _new_name: Optional[str] = field(metadata=config(field_name="newName"))
        """The new name for the catalog."""

        def __init__(self, new_name: str):
            super().__init__("rename")
            self._new_name = new_name

        def catalog_change(self):
            return CatalogChange.rename(self._new_name)

        def validate(self):
            """Validates the fields of the request.

            Raises:
                IllegalArgumentException if the new name is not set.
            """
            if not self._new_name:
                raise ValueError('"newName" field is required and cannot be empty')

    @dataclass
    class UpdateCatalogCommentRequest(CatalogUpdateRequestBase):
        """Request to update the comment of a catalog."""

        _new_comment: Optional[str] = field(metadata=config(field_name="newComment"))
        """The new comment for the catalog."""

        def __init__(self, new_comment: str):
            super().__init__("updateComment")
            self._new_comment = new_comment

        def catalog_change(self):
            return CatalogChange.update_comment(self._new_comment)

        def validate(self):
            """Validates the fields of the request. Always pass."""
            pass

    @dataclass
    class SetCatalogPropertyRequest(CatalogUpdateRequestBase):
        """Request to set a property on a catalog."""

        _property: Optional[str] = field(metadata=config(field_name="property"))
        """The property to set."""

        _value: Optional[str] = field(metadata=config(field_name="value"))
        """The value of the property."""

        def __init__(self, catalog_property: str, value: str):
            super().__init__("setProperty")
            self._property = catalog_property
            self._value = value

        def catalog_change(self):
            return CatalogChange.set_property(self._property, self._value)

        def validate(self):
            if not self._property:
                raise ValueError('"property" field is required and cannot be empty')
            if not self._value:
                raise ValueError('"value" field is required and cannot be empty')

    class RemoveCatalogPropertyRequest(CatalogUpdateRequestBase):
        """Request to remove a property from a catalog."""

        property: Optional[str] = None
        """The property to remove."""

        def __init__(self, catalog_property: str):
            super().__init__("removeProperty")
            self._property = catalog_property

        def catalog_change(self):
            return CatalogChange.remove_property(self._property)

        def validate(self):
            if not self._property:
                raise ValueError('"property" field is required and cannot be empty')



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import time
import json
import base64
from typing import Optional
from gravitino.auth.oauth2_token_provider import OAuth2TokenProvider
from gravitino.dto.responses.oauth2_token_response import OAuth2TokenResponse
from gravitino.dto.requests.oauth2_client_credential_request import (
    OAuth2ClientCredentialRequest,
)
from gravitino.exceptions.base import (
    GravitinoRuntimeException,
    IllegalArgumentException,
)
from gravitino.exceptions.handlers.oauth_error_handler import OAUTH_ERROR_HANDLER

CLIENT_CREDENTIALS = "client_credentials"
CREDENTIAL_SPLITTER = ":"
TOKEN_SPLITTER = "."
JWT_EXPIRE = "exp"


class DefaultOAuth2TokenProvider(OAuth2TokenProvider):
    """This class is the default implement of OAuth2TokenProvider."""

    _credential: Optional[str]
    _scope: Optional[str]
    _path: Optional[str]
    _token: Optional[str]

    def __init__(
        self,
        uri: str = None,
        credential: str = None,
        scope: str = None,
        path: str = None,
    ):
        super().__init__(uri)

        self._credential = credential
        self._scope = scope
        self._path = path

        self.validate()

        self._token = self._fetch_token()

    def validate(self):
        if not self._credential or not self._credential.strip():
            raise IllegalArgumentException("OAuth2TokenProvider must set credential")

        if not self._scope or not self._scope.strip():
            raise IllegalArgumentException("OAuth2TokenProvider must set scope")

        if not self._path or not self._path.strip():
            raise IllegalArgumentException("OAuth2TokenProvider must set path")

    def _get_access_token(self) -> Optional[str]:

        expires = self._expires_at_millis()

        if expires is None:
            return None

        if expires > time.time() * 1000:
            return self._token

        self._token = self._fetch_token()
        return self._token

    def _parse_credential(self):
        if self._credential is None:
            raise ValueError("Invalid credential: None")

        credential_info = self._credential.split(CREDENTIAL_SPLITTER, maxsplit=1)
        client_id = None
        client_secret = None

        if len(credential_info) == 2:
            client_id, client_secret = credential_info
        elif len(credential_info) == 1:
            client_secret = credential_info[0]
        else:
            raise GravitinoRuntimeException(f"Invalid credential: {self._credential}")

        return client_id, client_secret

    def _fetch_token(self) -> str:

        client_id, client_secret = self._parse_credential()

        client_credential_request = OAuth2ClientCredentialRequest(
            grant_type=CLIENT_CREDENTIALS,
            client_id=client_id,
            client_secret=client_secret,
            scope=self._scope,
        )

        resp = self._client.post_form(
            self._path,
            data=client_credential_request,
            error_handler=OAUTH_ERROR_HANDLER,
        )
        oauth2_resp = OAuth2TokenResponse.from_json(resp.body, infer_missing=True)
        oauth2_resp.validate()

        return oauth2_resp.access_token()

    def _expires_at_millis(self) -> int:
        if self._token is None:
            return None

        parts = self._token.split(TOKEN_SPLITTER)

        if len(parts) != 3:
            return None

        jwt = json.loads(
            base64.b64decode(parts[1] + "=" * (-len(parts[1]) % 4)).decode("utf-8")
        )

        if JWT_EXPIRE not in jwt or not isinstance(jwt[JWT_EXPIRE], int):
            return None

        return jwt[JWT_EXPIRE] * 1000



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import logging
from typing import List, Dict

from gravitino.api.catalog import Catalog
from gravitino.api.credential.supports_credentials import SupportsCredentials
from gravitino.api.credential.credential import Credential
from gravitino.api.fileset import Fileset
from gravitino.api.fileset_change import FilesetChange
from gravitino.audit.caller_context import CallerContextHolder, CallerContext
from gravitino.client.base_schema_catalog import BaseSchemaCatalog
from gravitino.client.generic_fileset import GenericFileset
from gravitino.dto.audit_dto import AuditDTO
from gravitino.dto.requests.fileset_create_request import FilesetCreateRequest
from gravitino.dto.requests.fileset_update_request import FilesetUpdateRequest
from gravitino.dto.requests.fileset_updates_request import FilesetUpdatesRequest
from gravitino.dto.responses.drop_response import DropResponse
from gravitino.dto.responses.entity_list_response import EntityListResponse
from gravitino.dto.responses.file_location_response import FileLocationResponse
from gravitino.dto.responses.fileset_response import FilesetResponse
from gravitino.name_identifier import NameIdentifier
from gravitino.namespace import Namespace
from gravitino.utils import HTTPClient
from gravitino.rest.rest_utils import encode_string
from gravitino.exceptions.handlers.fileset_error_handler import FILESET_ERROR_HANDLER

logger = logging.getLogger(__name__)


class FilesetCatalog(BaseSchemaCatalog, SupportsCredentials):
    """
    Fileset catalog is a catalog implementation that supports fileset like metadata operations, for
    example, schemas and filesets list, creation, update and deletion. A Fileset catalog is under the metalake.
    """

    def __init__(
        self,
        namespace: Namespace,
        name: str = None,
        catalog_type: Catalog.Type = Catalog.Type.UNSUPPORTED,
        provider: str = None,
        comment: str = None,
        properties: Dict[str, str] = None,
        audit: AuditDTO = None,
        rest_client: HTTPClient = None,
    ):

        super().__init__(
            namespace,
            name,
            catalog_type,
            provider,
            comment,
            properties,
            audit,
            rest_client,
        )

    def as_fileset_catalog(self):
        return self

    def list_filesets(self, namespace: Namespace) -> List[NameIdentifier]:
        """List the filesets in a schema namespace from the catalog.

        Args:
            namespace: A schema namespace. This namespace should have 1 level, which is the schema name

        Raises:
            NoSuchSchemaException If the schema does not exist.

        Returns:
            A list of NameIdentifier of filesets under the given namespace.
        """

        self.check_fileset_namespace(namespace)

        full_namespace = self._get_fileset_full_namespace(namespace)

        resp = self.rest_client.get(
            self.format_fileset_request_path(full_namespace),
            error_handler=FILESET_ERROR_HANDLER,
        )
        entity_list_resp = EntityListResponse.from_json(resp.body, infer_missing=True)
        entity_list_resp.validate()

        return [
            NameIdentifier.of(ident.namespace().level(2), ident.name())
            for ident in entity_list_resp.identifiers()
        ]

    def load_fileset(self, ident: NameIdentifier) -> Fileset:
        """Load fileset metadata by {@link NameIdentifier} from the catalog.

        Args:
            ident: A fileset identifier, which should be "schema.fileset" format.

        Raises:
            NoSuchFilesetException If the fileset does not exist.

        Returns:
            The fileset metadata.
        """
        self.check_fileset_name_identifier(ident)

        full_namespace = self._get_fileset_full_namespace(ident.namespace())

        resp = self.rest_client.get(
            f"{self.format_fileset_request_path(full_namespace)}/{encode_string(ident.name())}",
            error_handler=FILESET_ERROR_HANDLER,
        )
        fileset_resp = FilesetResponse.from_json(resp.body, infer_missing=True)
        fileset_resp.validate()

        return GenericFileset(fileset_resp.fileset(), self.rest_client, full_namespace)

    def create_fileset(
        self,
        ident: NameIdentifier,
        comment: str,
        fileset_type: Fileset.Type,
        storage_location: str,
        properties: Dict[str, str],
    ) -> Fileset:
        """Create a fileset metadata in the catalog.

        If the type of the fileset object is "MANAGED", the underlying storageLocation can be null,
        and Gravitino will manage the storage location based on the location of the schema.

        If the type of the fileset object is "EXTERNAL", the underlying storageLocation must be set.

        Args:
            ident: A fileset identifier, which should be "schema.fileset" format.
            comment: The comment of the fileset.
            fileset_type: The type of the fileset.
            storage_location: The storage location of the fileset.
            properties: The properties of the fileset.

        Raises:
            NoSuchSchemaException If the schema does not exist.
            FilesetAlreadyExistsException If the fileset already exists.

        Returns:
            The created fileset metadata
        """
        self.check_fileset_name_identifier(ident)

        full_namespace = self._get_fileset_full_namespace(ident.namespace())

        req = FilesetCreateRequest(
            name=encode_string(ident.name()),
            comment=comment,
            fileset_type=fileset_type,
            storage_location=storage_location,
            properties=properties,
        )

        resp = self.rest_client.post(
            self.format_fileset_request_path(full_namespace),
            req,
            error_handler=FILESET_ERROR_HANDLER,
        )
        fileset_resp = FilesetResponse.from_json(resp.body, infer_missing=True)
        fileset_resp.validate()

        return fileset_resp.fileset()

    def alter_fileset(self, ident: NameIdentifier, *changes) -> Fileset:
        """Update a fileset metadata in the catalog.

        Args:
            ident: A fileset identifier, which should be "schema.fileset" format.
            changes: The changes to apply to the fileset.

        Args:
            IllegalArgumentException If the changes are invalid.
            NoSuchFilesetException If the fileset does not exist.

        Returns:
            The updated fileset metadata.
        """
        self.check_fileset_name_identifier(ident)

        full_namespace = self._get_fileset_full_namespace(ident.namespace())

        updates = [
            FilesetCatalog.to_fileset_update_request(change) for change in changes
        ]
        req = FilesetUpdatesRequest(updates)
        req.validate()

        resp = self.rest_client.put(
            f"{self.format_fileset_request_path(full_namespace)}/{encode_string(ident.name())}",
            req,
            error_handler=FILESET_ERROR_HANDLER,
        )
        fileset_resp = FilesetResponse.from_json(resp.body, infer_missing=True)
        fileset_resp.validate()

        return fileset_resp.fileset()

    def drop_fileset(self, ident: NameIdentifier) -> bool:
        """Drop a fileset from the catalog.

        The underlying files will be deleted if this fileset type is managed, otherwise, only the
        metadata will be dropped.

        Args:
             ident: A fileset identifier, which should be "schema.fileset" format.

        Returns:
             true If the fileset is dropped, false the fileset did not exist.
        """
        self.check_fileset_name_identifier(ident)

        full_namespace = self._get_fileset_full_namespace(ident.namespace())

        resp = self.rest_client.delete(
            f"{self.format_fileset_request_path(full_namespace)}/{encode_string(ident.name())}",
            error_handler=FILESET_ERROR_HANDLER,
        )
        drop_resp = DropResponse.from_json(resp.body, infer_missing=True)
        drop_resp.validate()

        return drop_resp.dropped()

    def get_file_location(self, ident: NameIdentifier, sub_path: str) -> str:
        """Get the actual location of a file or directory based on the storage location of Fileset and the sub path.

        Args:
             ident: A fileset identifier, which should be "schema.fileset" format.
             sub_path: The sub path of the file or directory.

        Returns:
             The actual location of the file or directory.
        """
        self.check_fileset_name_identifier(ident)

        full_namespace = self._get_fileset_full_namespace(ident.namespace())
        try:
            caller_context: CallerContext = CallerContextHolder.get()
            params = {"sub_path": encode_string(sub_path)}

            resp = self.rest_client.get(
                self.format_file_location_request_path(full_namespace, ident.name()),
                params=params,
                headers=(
                    caller_context.context() if caller_context is not None else None
                ),
                error_handler=FILESET_ERROR_HANDLER,
            )
            file_location_resp = FileLocationResponse.from_json(
                resp.body, infer_missing=True
            )
            file_location_resp.validate()

            return file_location_resp.file_location()
        finally:
            CallerContextHolder.remove()

    @staticmethod
    def check_fileset_namespace(namespace: Namespace):
        Namespace.check(
            namespace is not None and namespace.length() == 1,
            f"Fileset namespace must be non-null and have 1 level, the input namespace is {namespace}",
        )

    @staticmethod
    def check_fileset_name_identifier(ident: NameIdentifier):
        NameIdentifier.check(ident is not None, "NameIdentifier must not be None")
        NameIdentifier.check(
            ident.name() is not None and len(ident.name()) != 0,
            "NameIdentifier name must not be empty",
        )
        FilesetCatalog.check_fileset_namespace(ident.namespace())

    def _get_fileset_full_namespace(self, fileset_namespace: Namespace) -> Namespace:
        return Namespace.of(
            self._catalog_namespace.level(0), self.name(), fileset_namespace.level(0)
        )

    @staticmethod
    def format_fileset_request_path(namespace: Namespace) -> str:
        schema_ns = Namespace.of(namespace.level(0), namespace.level(1))
        return f"{BaseSchemaCatalog.format_schema_request_path(schema_ns)}/{encode_string(namespace.level(2))}/filesets"

    @staticmethod
    def format_file_location_request_path(namespace: Namespace, name: str) -> str:
        schema_ns = Namespace.of(namespace.level(0), namespace.level(1))
        return (
            f"{BaseSchemaCatalog.format_schema_request_path(schema_ns)}/{encode_string(namespace.level(2))}"
            f"/filesets/{encode_string(name)}/location"
        )

    @staticmethod
    def to_fileset_update_request(change: FilesetChange):
        if isinstance(change, FilesetChange.RenameFileset):
            return FilesetUpdateRequest.RenameFilesetRequest(change.new_name())
        if isinstance(change, FilesetChange.UpdateFilesetComment):
            return FilesetUpdateRequest.UpdateFilesetCommentRequest(
                change.new_comment()
            )
        if isinstance(change, FilesetChange.SetProperty):
            return FilesetUpdateRequest.SetFilesetPropertyRequest(
                change.property(), change.value()
            )
        if isinstance(change, FilesetChange.RemoveProperty):
            return FilesetUpdateRequest.RemoveFilesetPropertyRequest(change.property())
        if isinstance(change, FilesetChange.RemoveComment):
            return FilesetUpdateRequest.UpdateFilesetCommentRequest(None)
        raise ValueError(f"Unknown change type: {type(change).__name__}")

    def support_credentials(self) -> SupportsCredentials:
        return self

    def get_credentials(self) -> List[Credential]:
        return self._object_credential_operations.get_credentials()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import Dict

from gravitino.api.credential.credential import Credential
from gravitino.api.credential.gcs_token_credential import GCSTokenCredential
from gravitino.api.credential.oss_token_credential import OSSTokenCredential
from gravitino.api.credential.s3_secret_key_credential import S3SecretKeyCredential
from gravitino.api.credential.s3_token_credential import S3TokenCredential
from gravitino.api.credential.oss_secret_key_credential import OSSSecretKeyCredential
from gravitino.api.credential.adls_token_credential import ADLSTokenCredential
from gravitino.api.credential.azure_account_key_credential import (
    AzureAccountKeyCredential,
)


class CredentialFactory:
    @staticmethod
    def create(
        credential_type: str, credential_info: Dict[str, str], expire_time_in_ms: int
    ) -> Credential:
        credential = None

        if credential_type == S3TokenCredential.S3_TOKEN_CREDENTIAL_TYPE:
            credential = S3TokenCredential(credential_info, expire_time_in_ms)
        elif credential_type == S3SecretKeyCredential.S3_SECRET_KEY_CREDENTIAL_TYPE:
            credential = S3SecretKeyCredential(credential_info, expire_time_in_ms)
        elif credential_type == GCSTokenCredential.GCS_TOKEN_CREDENTIAL_TYPE:
            credential = GCSTokenCredential(credential_info, expire_time_in_ms)
        elif credential_type == OSSTokenCredential.OSS_TOKEN_CREDENTIAL_TYPE:
            credential = OSSTokenCredential(credential_info, expire_time_in_ms)
        elif credential_type == OSSSecretKeyCredential.OSS_SECRET_KEY_CREDENTIAL_TYPE:
            credential = OSSSecretKeyCredential(credential_info, expire_time_in_ms)
        elif credential_type == ADLSTokenCredential.ADLS_TOKEN_CREDENTIAL_TYPE:
            credential = ADLSTokenCredential(credential_info, expire_time_in_ms)
        elif (
            credential_type
            == AzureAccountKeyCredential.AZURE_ACCOUNT_KEY_CREDENTIAL_TYPE
        ):
            credential = AzureAccountKeyCredential(credential_info, expire_time_in_ms)
        else:
            raise NotImplementedError(
                f"Credential type {credential_type} is not supported"
            )

        return credential



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import ABC
from typing import Dict

from gravitino.api.credential.credential import Credential
from gravitino.utils.precondition import Precondition


class OSSSecretKeyCredential(Credential, ABC):
    """Represents OSS secret key credential."""

    OSS_SECRET_KEY_CREDENTIAL_TYPE: str = "oss-secret-key"
    _STATIC_ACCESS_KEY_ID: str = "oss-access-key-id"
    _STATIC_SECRET_ACCESS_KEY: str = "oss-secret-access-key"

    def __init__(self, credential_info: Dict[str, str], expire_time_in_ms: int):
        self._access_key_id = credential_info.get(self._STATIC_ACCESS_KEY_ID, None)
        self._secret_access_key = credential_info.get(
            self._STATIC_SECRET_ACCESS_KEY, None
        )
        Precondition.check_string_not_empty(
            self._access_key_id, "The OSS access key ID should not be empty"
        )
        Precondition.check_string_not_empty(
            self._secret_access_key, "The OSS secret access key should not be empty"
        )
        Precondition.check_argument(
            expire_time_in_ms == 0,
            "The expiration time of OSS secret key credential should be 0",
        )

    def credential_type(self) -> str:
        """Returns the type of the credential.

        Returns:
            The type of the credential.
        """
        return self.OSS_SECRET_KEY_CREDENTIAL_TYPE

    def expire_time_in_ms(self) -> int:
        """Returns the expiration time of the credential in milliseconds since
        the epoch, 0 means it will never expire.

        Returns:
            The expiration time of the credential.
        """
        return 0

    def credential_info(self) -> Dict[str, str]:
        """The credential information.

        Returns:
            The credential information.
        """
        return {
            self._STATIC_ACCESS_KEY_ID: self._access_key_id,
            self._STATIC_SECRET_ACCESS_KEY: self._secret_access_key,
        }

    def access_key_id(self) -> str:
        """The OSS access key ID.

        Returns:
            The OSS access key ID.
        """
        return self._access_key_id

    def secret_access_key(self) -> str:
        """The OSS secret access key.

        Returns:
            The OSS secret access key.
        """
        return self._secret_access_key



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import ClassVar

from dataclasses import dataclass, field

from dataclasses_json import DataClassJsonMixin, config

from gravitino.exceptions.base import (
    IllegalNameIdentifierException,
)
from gravitino.namespace import Namespace


@dataclass
class NameIdentifier(DataClassJsonMixin):
    """A name identifier is a sequence of names separated by dots. It's used to identify a metalake, a
    catalog, a schema or a table. For example, "metalake1" can represent a metalake,
    "metalake1.catalog1" can represent a catalog, "metalake1.catalog1.schema1" can represent a
    schema.
    """

    _DOT: ClassVar[str] = "."

    _name: str = field(metadata=config(field_name="name"))
    _namespace: Namespace = field(
        metadata=config(
            field_name="namespace",
            encoder=Namespace.to_json,
            decoder=Namespace.from_json,
        )
    )

    @classmethod
    def builder(cls, namespace: Namespace, name: str):
        return NameIdentifier(_namespace=namespace, _name=name)

    def namespace(self):
        return self._namespace

    def name(self):
        return self._name

    @staticmethod
    def of(*names: str) -> "NameIdentifier":
        """Create the NameIdentifier with the given levels of names.

        Args:
            names The names of the identifier

        Returns:
            The created NameIdentifier
        """

        NameIdentifier.check(
            names is not None, "Cannot create a NameIdentifier with null names"
        )
        NameIdentifier.check(
            len(names) > 0, "Cannot create a NameIdentifier with no names"
        )

        return NameIdentifier.builder(Namespace.of(*names[:-1]), names[-1])

    @staticmethod
    def parse(identifier: str) -> "NameIdentifier":
        """Create a NameIdentifier from the given identifier string.

        Args:
            identifier: The identifier string

        Returns:
            The created NameIdentifier
        """
        NameIdentifier.check(
            identifier is not None and identifier != "",
            "Cannot parse a null or empty identifier",
        )

        parts = identifier.split(NameIdentifier._DOT)
        return NameIdentifier.of(*parts)

    def has_namespace(self):
        """Check if the NameIdentifier has a namespace.

        Returns:
            True if the NameIdentifier has a namespace, false otherwise.
        """
        return not self.namespace().is_empty()

    def get_namespace(self):
        """Get the namespace of the NameIdentifier.

        Returns:
            The namespace of the NameIdentifier.
        """
        return self._namespace

    def get_name(self):
        """Get the name of the NameIdentifier.

        Returns:
            The name of the NameIdentifier.
        """
        return self._name

    def __eq__(self, other):
        if not isinstance(other, NameIdentifier):
            return False
        return self._namespace == other._namespace and self._name == other._name

    def __hash__(self):
        return hash((self._namespace, self._name))

    def __str__(self):
        if self.has_namespace():
            return str(self._namespace) + "." + self._name
        return self._name

    @staticmethod
    def check(condition, message, *args):
        """Check the given condition is true. Throw an {@link IllegalNameIdentifierException} if it's not.

        Args:
            condition: The condition to check.
            message: The message to throw.
            args: The arguments to the message.
        """
        if not condition:
            raise IllegalNameIdentifierException(message.format(*args))



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import ABC
from dataclasses import dataclass, field

from dataclasses_json import config


class FilesetChange(ABC):
    """A fileset change is a change to a fileset. It can be used to rename a fileset, update the comment
    of a fileset, set a property and value pair for a fileset, or remove a property from a fileset.
    """

    @staticmethod
    def rename(new_name):
        """Creates a new fileset change to rename the fileset.

        Args:
            new_name: The new name of the fileset.

        Returns:
             The fileset change.
        """
        return FilesetChange.RenameFileset(new_name)

    @staticmethod
    def update_comment(new_comment):
        """Creates a new fileset change to update the fileset comment.

        Args:
            new_comment: The new comment for the fileset.

        Returns:
             The fileset change.
        """
        return FilesetChange.UpdateFilesetComment(new_comment)

    @staticmethod
    def set_property(fileset_property, value):
        """Creates a new fileset change to set the property and value for the fileset.

        Args:
            fileset_property: The property name to set.
            value: The value to set the property to.

        Returns:
             The fileset change.
        """
        return FilesetChange.SetProperty(fileset_property, value)

    @staticmethod
    def remove_property(fileset_property):
        """Creates a new fileset change to remove a property from the fileset.

        Args:
            fileset_property: The property name to remove.

        Returns:
            The fileset change.
        """
        return FilesetChange.RemoveProperty(fileset_property)

    @staticmethod
    def remove_comment():
        """Creates a new fileset change to remove comment from the fileset.

        Returns:
            The fileset change.

        Deprecated:
            Please use `update_comment(str)` with null value as the argument instead.
        """
        return FilesetChange.UpdateFilesetComment(None)

    @dataclass
    class RenameFileset:
        """A fileset change to rename the fileset."""

        _new_name: str = field(metadata=config(field_name="new_name"))

        def new_name(self):
            """Retrieves the new name set for the fileset.

            Returns:
                 The new name of the fileset.
            """
            return self._new_name

        def __eq__(self, other):
            """Compares this RenameFileset instance with another object for equality.
            Two instances are considered equal if they designate the same new name for the fileset.

            Args:
                 other: The object to compare with this instance.

            Returns:
                 true if the given object represents an identical fileset renaming operation; false otherwise.
            """
            if not isinstance(other, FilesetChange.RenameFileset):
                return False
            return self._new_name == other.new_name()

        def __hash__(self):
            """Generates a hash code for this RenameFileset instance.
            The hash code is primarily based on the new name for the fileset.

            Returns:
                 A hash code value for this renaming operation.
            """
            return hash(self._new_name)

        def __str__(self):
            """Provides a string representation of the RenameFile instance.
            This string includes the class name followed by the new name of the fileset.

            Returns:
                 A string summary of this renaming operation.
            """
            return f"RENAMEFILESET {self._new_name}"

    @dataclass
    class UpdateFilesetComment:
        """A fileset change to update the fileset comment."""

        _new_comment: str = field(metadata=config(field_name="new_comment"))

        def new_comment(self):
            """Retrieves the new comment intended for the fileset.

            Returns:
                 The new comment that has been set for the fileset.
            """
            return self._new_comment

        def __eq__(self, other) -> bool:
            """Compares this UpdateFilesetComment instance with another object for equality.
            Two instances are considered equal if they designate the same new comment for the fileset.

            Args:
                 other: The object to compare with this instance.

            Returns:
                 true if the given object represents the same comment update; false otherwise.
            """
            if not isinstance(other, FilesetChange.UpdateFilesetComment):
                return False
            return self._new_comment == other.new_comment()

        def __hash__(self):
            """Generates a hash code for this UpdateFileComment instance.
            The hash code is based on the new comment for the fileset.

            Returns:
                 A hash code representing this comment update operation.
            """
            return hash(self._new_comment)

        def __str__(self):
            """Provides a string representation of the UpdateFilesetComment instance.
            This string format includes the class name followed by the new comment for the fileset.

            Returns:
                 A string summary of this comment update operation.
            """
            return f"UPDATEFILESETCOMMENT {self._new_comment}"

    @dataclass
    class SetProperty:
        """A fileset change to set the property and value for the fileset."""

        _property: str = field(metadata=config(field_name="property"))
        _value: str = field(metadata=config(field_name="value"))

        def property(self):
            """Retrieves the name of the property being set in the fileset.

            Returns:
                 The name of the property.
            """
            return self._property

        def value(self):
            """Retrieves the value assigned to the property in the fileset.

            Returns:
                 The value of the property.
            """
            return self._value

        def __eq__(self, other) -> bool:
            """Compares this SetProperty instance with another object for equality.
            Two instances are considered equal if they have the same property and value for the fileset.

            Args:
                 other: The object to compare with this instance.

            Returns:
                 true if the given object represents the same property setting; false otherwise.
            """
            if not isinstance(other, FilesetChange.SetProperty):
                return False
            return self._property == other.property() and self._value == other.value()

        def __hash__(self):
            """Generates a hash code for this SetProperty instance.
            The hash code is based on both the property name and its assigned value.

            Returns:
                 A hash code value for this property setting.
            """
            return hash((self._property, self._value))

        def __str__(self):
            """Provides a string representation of the SetProperty instance.
            This string format includes the class name followed by the property and its value.

            Returns:
                 A string summary of the property setting.
            """
            return f"SETPROPERTY {self._property} {self._value}"

    @dataclass
    class RemoveProperty:
        """A fileset change to remove a property from the fileset."""

        _property: str = field(metadata=config(field_name="property"))

        def property(self):
            """Retrieves the name of the property to be removed from the fileset.

            Returns:
                 The name of the property for removal.
            """
            return self._property

        def __eq__(self, other) -> bool:
            """Compares this RemoveProperty instance with another object for equality.
            Two instances are considered equal if they target the same property for removal from the fileset.

            Args:
                 other: The object to compare with this instance.

            Returns:
                 true if the given object represents the same property removal; false otherwise.
            """
            if not isinstance(other, FilesetChange.RemoveProperty):
                return False
            return self._property == other.property()

        def __hash__(self):
            """Generates a hash code for this RemoveProperty instance.
            The hash code is based on the property name that is to be removed from the fileset.

            Returns:
                 A hash code value for this property removal operation.
            """
            return hash(self._property)

        def __str__(self):
            """Provides a string representation of the RemoveProperty instance.
            This string format includes the class name followed by the property name to be removed.

            Returns:
                 A string summary of the property removal operation.
            """
            return f"REMOVEPROPERTY {self._property}"

    @dataclass
    class RemoveComment:
        """A fileset change to remove comment from the fileset.

        Deprecated:
            Please use `UpdateFilesetComment(str)` with null value as the argument instead.
        """

        def __eq__(self, other) -> bool:
            """Compares this RemoveComment instance with another object for equality.
            Two instances are considered equal if they are RemoveComment instance.

            Args:
                 other: The object to compare with this instance.

            Returns:
                 true if the given object represents the comment removal; false otherwise.
            """
            return isinstance(other, FilesetChange.RemoveComment)

        def __hash__(self):
            """Generates a hash code for this RemoveComment instance.
            The hash code is based on the RemoveComment instance name.

            Returns:
                 A hash code value for comment removal operation.
            """
            return hash("REMOVECOMMENT")

        def __str__(self):
            """Provides a string representation of the RemoveComment instance.
            This string format includes the class name.

            Returns:
                 A string summary of the comment removal operation.
            """
            return "REMOVECOMMENT"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from gravitino.api.catalog import Catalog
from gravitino.api.schema import Schema
from gravitino.api.fileset import Fileset
from gravitino.api.fileset_change import FilesetChange
from gravitino.api.metalake_change import MetalakeChange
from gravitino.api.schema_change import SchemaChange
from gravitino.client.gravitino_client import GravitinoClient
from gravitino.client.gravitino_admin_client import GravitinoAdminClient
from gravitino.client.gravitino_metalake import GravitinoMetalake
from gravitino.name_identifier import NameIdentifier
from gravitino.filesystem import gvfs



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from gravitino.api.catalog import Catalog
from gravitino.api.catalog_change import CatalogChange
from gravitino.client.fileset_catalog import FilesetCatalog
from gravitino.client.generic_model_catalog import GenericModelCatalog
from gravitino.dto.catalog_dto import CatalogDTO
from gravitino.dto.requests.catalog_update_request import CatalogUpdateRequest
from gravitino.dto.requests.metalake_update_request import MetalakeUpdateRequest
from gravitino.api.metalake_change import MetalakeChange
from gravitino.utils import HTTPClient
from gravitino.namespace import Namespace


class DTOConverters:
    """Utility class for converting between DTOs and domain objects."""

    @staticmethod
    def to_metalake_update_request(change: MetalakeChange) -> object:
        # Assuming MetalakeUpdateRequest has similar nested class structure for requests
        if isinstance(change, MetalakeChange.RenameMetalake):
            return MetalakeUpdateRequest.RenameMetalakeRequest(change.new_name())
        if isinstance(change, MetalakeChange.UpdateMetalakeComment):
            return MetalakeUpdateRequest.UpdateMetalakeCommentRequest(
                change.new_comment()
            )
        if isinstance(change, MetalakeChange.SetProperty):
            return MetalakeUpdateRequest.SetMetalakePropertyRequest(
                change.property(), change.value()
            )
        if isinstance(change, MetalakeChange.RemoveProperty):
            return MetalakeUpdateRequest.RemoveMetalakePropertyRequest(
                change.property()
            )

        raise ValueError(f"Unknown change type: {type(change).__name__}")

    @staticmethod
    def to_catalog(metalake: str, catalog: CatalogDTO, client: HTTPClient):
        namespace = Namespace.of(metalake)
        if catalog.type() == Catalog.Type.FILESET:
            return FilesetCatalog(
                namespace=namespace,
                name=catalog.name(),
                catalog_type=catalog.type(),
                provider=catalog.provider(),
                comment=catalog.comment(),
                properties=catalog.properties(),
                audit=catalog.audit_info(),
                rest_client=client,
            )

        if catalog.type() == Catalog.Type.MODEL:
            return GenericModelCatalog(
                namespace=namespace,
                name=catalog.name(),
                catalog_type=catalog.type(),
                provider=catalog.provider(),
                comment=catalog.comment(),
                properties=catalog.properties(),
                audit=catalog.audit_info(),
                rest_client=client,
            )

        raise NotImplementedError("Unsupported catalog type: " + str(catalog.type()))

    @staticmethod
    def to_catalog_update_request(change: CatalogChange):
        if isinstance(change, CatalogChange.RenameCatalog):
            return CatalogUpdateRequest.RenameCatalogRequest(change.new_name())
        if isinstance(change, CatalogChange.UpdateCatalogComment):
            return CatalogUpdateRequest.UpdateCatalogCommentRequest(
                change.new_comment()
            )
        if isinstance(change, CatalogChange.SetProperty):
            # TODO
            # pylint: disable=too-many-function-args
            return CatalogUpdateRequest.SetCatalogPropertyRequest(
                change.property(), change.value()
            )
        if isinstance(change, CatalogChange.RemoveProperty):
            return CatalogUpdateRequest.RemoveCatalogPropertyRequest(
                change.get_property()
            )

        raise ValueError(f"Unknown change type: {type(change).__name__}")



"""
MIT License

Copyright (c) 2016 Dhamu

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import logging
from typing import Tuple
from urllib.request import Request, build_opener
from urllib.parse import urlencode

from urllib.error import HTTPError
import json as _json

from gravitino.auth.auth_constants import AuthConstants
from gravitino.auth.auth_data_provider import AuthDataProvider
from gravitino.typing import JSONType

from gravitino.constants.timeout import TIMEOUT

from gravitino.dto.responses.error_response import ErrorResponse
from gravitino.dto.responses.oauth2_error_response import OAuth2ErrorResponse
from gravitino.exceptions.base import RESTException, UnknownError
from gravitino.exceptions.handlers.error_handler import ErrorHandler

logger = logging.getLogger(__name__)


class Response:
    def __init__(self, response):
        self._status_code = response.getcode()
        self._body = response.read()
        self._headers = response.info()
        self._url = response.url

        logging.basicConfig(level=logging.DEBUG)
        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.DEBUG)
        logger.addHandler(console_handler)

    @property
    def status_code(self):
        return self._status_code

    @property
    def url(self):
        return self._url

    @property
    def body(self):
        return self._body

    @property
    def headers(self):
        return self._headers

    def json(self):
        if self.body:
            return _json.loads(self.body.decode("utf-8"))
        return None


class HTTPClient:

    FORMDATA_HEADER = {
        "Content-Type": "application/x-www-form-urlencoded",
        "Accept": "application/vnd.gravitino.v1+json",
    }

    JSON_HEADER = {
        "Content-Type": "application/json",
        "Accept": "application/vnd.gravitino.v1+json",
    }

    def __init__(
        self,
        host,
        *,
        request_headers=None,
        timeout=TIMEOUT,
        is_debug=False,
        auth_data_provider: AuthDataProvider = None,
    ) -> None:
        self.host = host
        self.request_headers = request_headers or {}
        self.timeout = timeout
        self.is_debug = is_debug
        self.auth_data_provider = auth_data_provider

    def _build_url(self, endpoint=None, params=None):
        url = self.host

        if endpoint:
            url = f"{url.rstrip('/')}/{endpoint.lstrip('/')}"

        if params:
            params = {k: v for k, v in params.items() if v is not None}
            url_values = urlencode(sorted(params.items()), True)
            url = f"{url}?{url_values}"

        return url

    def _update_headers(self, request_headers):
        self.request_headers.update(request_headers)

    def _mask_auth_headers(self, headers):
        if self.is_debug:
            return headers

        _headers = {}
        for key, value in headers.items():
            if key.lower() == "authorization":
                _headers[key] = "******"
            else:
                _headers[key] = value
        return _headers

    def _make_request(self, opener, request, timeout=None) -> Tuple[bool, Response]:
        timeout = timeout or self.timeout
        try:
            return (True, Response(opener.open(request, timeout=timeout)))
        except HTTPError as err:
            err_body = err.read()

            if err_body is None or len(err_body) == 0:
                return (
                    False,
                    ErrorResponse.generate_error_response(RESTException, err.reason),
                )

            err_resp = self._parse_error_response(err_body)
            err_resp.validate()

            return (False, err_resp)

    def _parse_error_response(self, err_body: bytes) -> ErrorResponse:
        json_err_body = _json.loads(err_body)

        if "code" in json_err_body:
            return ErrorResponse.from_json(err_body, infer_missing=True)

        return OAuth2ErrorResponse.from_json(err_body, infer_missing=True)

    # pylint: disable=too-many-locals
    def _request(
        self,
        method,
        endpoint,
        params=None,
        json=None,
        data=None,
        headers=None,
        timeout=None,
        error_handler: ErrorHandler = None,
    ):
        method = method.upper()
        request_data = None

        if data:
            request_data = urlencode(data.to_dict()).encode()
            self._update_headers(self.FORMDATA_HEADER)
        else:
            if json:
                request_data = json.to_json().encode("utf-8")

            self._update_headers(self.JSON_HEADER)

        if headers:
            self._update_headers(headers)

        opener = build_opener()
        request = Request(self._build_url(endpoint, params), data=request_data)
        if self.request_headers:
            for key, value in self.request_headers.items():
                request.add_header(key, value)
        if request_data and ("Content-Type" not in self.request_headers):
            request.add_header("Content-Type", "application/json")
        if self.auth_data_provider is not None:
            request.add_header(
                AuthConstants.HTTP_HEADER_AUTHORIZATION,
                self.auth_data_provider.get_token_data().decode("utf-8"),
            )
        request.get_method = lambda: method
        is_success, resp = self._make_request(opener, request, timeout=timeout)

        if is_success:
            return resp

        if not isinstance(error_handler, ErrorHandler):
            raise UnknownError(
                f"Unknown error handler {type(error_handler).__name__}, error response body: {resp}"
            ) from None

        error_handler.handle(resp)

        # This code generally will not be run because the error handler should define the default behavior,
        # and raise appropriate
        raise UnknownError(
            f"Error handler {type(error_handler).__name__} can't handle this response, error response body: {resp}"
        ) from None

    def get(self, endpoint, params=None, headers=None, error_handler=None, **kwargs):
        return self._request(
            "get",
            endpoint,
            params=params,
            headers=headers,
            error_handler=error_handler,
            **kwargs,
        )

    def delete(self, endpoint, error_handler=None, **kwargs):
        return self._request("delete", endpoint, error_handler=error_handler, **kwargs)

    def post(self, endpoint, json=None, error_handler=None, **kwargs):
        return self._request(
            "post", endpoint, json=json, error_handler=error_handler, **kwargs
        )

    def put(self, endpoint, json=None, error_handler=None, **kwargs):
        return self._request(
            "put", endpoint, json=json, error_handler=error_handler, **kwargs
        )

    def patch(self, endpoint, json=None, error_handler=None, **kwargs):
        return self._request(
            "patch", endpoint, json=json, error_handler=error_handler, **kwargs
        )

    def post_form(self, endpoint, data=None, error_handler=None, **kwargs):
        return self._request(
            "post", endpoint, data=data, error_handler=error_handler, **kwargs
        )

    def close(self):
        self._request("close", "/")
        if self.auth_data_provider is not None:
            self.auth_data_provider.close()


def unpack(path: str):
    def decorator(func):
        def wrapper(*args, **kwargs) -> JSONType:
            resp = func(*args, **kwargs)
            rv = resp.json()
            for p in path.split("."):
                if p not in rv:
                    raise KeyError(f"The path '{path}' can't find in dict")
                rv = rv.get(p)
            return rv

        return wrapper

    return decorator
